==> telsos-lib-0.1.462.jar: telsos/lib/pprint.clj <==
(ns telsos.lib.pprint
  (:require
   [lambdaisland.deep-diff2 :as ddiff]
   [telsos.lib.algorithms.seqs :as seqs]
   [telsos.lib.algorithms.trees :refer [->TreeEntry]])
  (:import
   (telsos.lib.algorithms.trees TreeEntry)))

(set! *warn-on-reflection*       true)
(set! *unchecked-math* :warn-on-boxed)

;; (PRETTY) PRINTING TREES
(def ^:private PPRINT-TREE-INDENT       "│   ")
(def ^:private PPRINT-TREE-EMPTYINDENT  "    ")
(def ^:private PPRINT-TREE-FORCHILD     "├── ")
(def ^:private PPRINT-TREE-FORLASTCHILD "└── ")
(def ^:private PPRINT-TREE-EOL          "\n")
(def ^:private PPRINT-TREE-EMPTY        "")

(defn- pprint-tree-genindent
  [[last? & last-child-infos]]
  (let [suffix (if last?
                 PPRINT-TREE-FORLASTCHILD
                 PPRINT-TREE-FORCHILD)
        prefix (->> last-child-infos
                    (butlast)
                    (reverse)
                    (map #(if % ;; empty?
                            PPRINT-TREE-EMPTYINDENT
                            PPRINT-TREE-INDENT))
                    (apply str))]

    (str prefix suffix)))

(defn- pprint-tree-impl
  [node adjs show depth level last-child-infos first?]
  (let [s     (show node)
        pfx   (if first? PPRINT-TREE-EMPTY PPRINT-TREE-EOL)
        level (long level)
        repr  (if (zero? level)
                (str pfx s)
                (str pfx (pprint-tree-genindent last-child-infos) s))]

    (print repr)

    (when-not (= level (long depth))
      (let [next-level (inc level)
            children   (adjs node)]
        (doseq [[child last?] (map vector children (seqs/mark-last children))]
          (pprint-tree-impl child adjs show depth next-level
                            #_last-child-infos (cons last? last-child-infos)

                            ;; 1st level only when called on top (in pprint-tree)
                            #_first? false))))))

(defn pprint-tree
  "Prints a tree using a textual representation like in UNIX tree command.
   adjs : node -> [node]
   show : node -> String"
  ([node adjs]
   (pprint-tree node adjs str))

  ([node adjs show]
   (pprint-tree node adjs show 9223372036854775807))

  ([node adjs show depth]
   (pprint-tree-impl node adjs show depth 0
                     #_last-child-infos '(true)
                     #_first? true)))

;; (PRETTY) PRINTING EDN TREES
(defn- edn-adjs
  [e]
  (cond
    (instance? TreeEntry e) [(:node e)]
    (set?                e) (seq e)
    (map?                e) (map (fn [[k v]] (->TreeEntry e k v))                 e)
    (sequential?         e) (map (fn  [i v]  (->TreeEntry e i v)) (iterate inc 0) e)

    :else  nil))

(defn- edn-repr
  [e]
  (cond (vector?     e) "[]"
        (list?       e) "(list)"
        (set?        e) "#{}"
        (sequential? e) "()"
        (record?     e) (str (type e))
        (map?        e) "{}"

        :else nil))

(defn- edn-show
  [e]
  (or (when (instance? TreeEntry e) (str (:index-or-key e)))
      (edn-repr e)
      (str e)))

(defn pprint-edn-tree
  ([e]
   (pprint-tree e edn-adjs edn-show))

  ([e depth]
   (pprint-tree e edn-adjs edn-show depth)))

;; (PRETTY) PRINTING DIFFS
(defn pprint-ddiff
  [x y]
  (ddiff/pretty-print (ddiff/diff x y)))

==> telsos-lib-0.1.468.jar: telsos/lib/pprint.clj <==
(ns telsos.lib.pprint
  (:require
   [clojure.string :as str]
   [lambdaisland.deep-diff2 :as ddiff]
   [telsos.lib.algorithms.seqs :as seqs]
   [telsos.lib.algorithms.trees :refer [->TreeEntry tree->associative-paths]])
  (:import
   (telsos.lib.algorithms.trees TreeEntry)))

(set! *warn-on-reflection*       true)
(set! *unchecked-math* :warn-on-boxed)

;; (PRETTY) PRINTING TREES
(def ^:private PPRINT-TREE-INDENT       "│   ")
(def ^:private PPRINT-TREE-EMPTYINDENT  "    ")
(def ^:private PPRINT-TREE-FORCHILD     "├── ")
(def ^:private PPRINT-TREE-FORLASTCHILD "└── ")
(def ^:private PPRINT-TREE-EOL          "\n")
(def ^:private PPRINT-TREE-EMPTY        "")

(defn- pprint-tree-genindent
  [[last? & last-child-infos]]
  (let [suffix (if last?
                 PPRINT-TREE-FORLASTCHILD
                 PPRINT-TREE-FORCHILD)
        prefix (->> last-child-infos
                    (butlast)
                    (reverse)
                    (map #(if % ;; empty?
                            PPRINT-TREE-EMPTYINDENT
                            PPRINT-TREE-INDENT))
                    (apply str))]

    (str prefix suffix)))

(defn- pprint-tree-impl
  [node adjs show depth level last-child-infos first?]
  (let [s     (show node)
        pfx   (if first? PPRINT-TREE-EMPTY PPRINT-TREE-EOL)
        level (long level)
        repr  (if (zero? level)
                (str pfx s)
                (str pfx (pprint-tree-genindent last-child-infos) s))]

    (print repr)

    (when-not (= level (long depth))
      (let [next-level (inc level)
            children   (adjs node)]
        (doseq [[child last?] (map vector children (seqs/mark-last children))]
          (pprint-tree-impl child adjs show depth next-level
                            #_last-child-infos (cons last? last-child-infos)

                            ;; 1st level only when called on top (in pprint-tree)
                            #_first? false))))))

(defn pprint-tree
  "Prints a tree using a textual representation like in UNIX tree command.
   adjs : node -> [node]
   show : node -> String"
  ([node adjs]
   (pprint-tree node adjs str))

  ([node adjs show]
   (pprint-tree node adjs show 9223372036854775807))

  ([node adjs show depth]
   (pprint-tree-impl node adjs show depth 0
                     #_last-child-infos '(true)
                     #_first? true)))

;; (PRETTY) PRINTING EDN (TREES AND PATHS)
(defn- edn-adjs
  [e]
  (cond
    (instance? TreeEntry e) [(:node e)]
    (set?                e) (seq e)
    (map?                e) (map (fn [[k v]] (->TreeEntry e k v))                 e)
    (sequential?         e) (map (fn  [i v]  (->TreeEntry e i v)) (iterate inc 0) e)
    :else  nil))

(defn- edn-repr
  [e]
  (cond (vector?     e) "["
        (list?       e) "("
        (set?        e) "#{"
        (sequential? e) "("
        (record?     e) (str (type e))
        (map?        e) "{"
        :else nil))

(defn double-quote-when-string [x]
  (if (string? x) (pr-str x) x))

(defn- edn-show
  [e]
  (or (when (instance? TreeEntry e) (double-quote-when-string (:index-or-key e)))
      (edn-repr e)
      (double-quote-when-string e)))

(defn pprint-edn-tree
  ([e]
   (pprint-tree e edn-adjs edn-show))

  ([e depth]
   (pprint-tree e edn-adjs edn-show depth)))

(defn- sort-edn-paths [paths]
  ;; Because of the way Clojure sorts vectors using their length improperly from our
  ;; needs perspective:
  (let [max-len (apply max (map count paths))]
    (sort-by #(vec (concat % (repeat (- max-len (count %)) Integer/MAX_VALUE)))
             paths)))

(def ^:private edn-complementary-symbol
  {(symbol "[")  (symbol "]")
   (symbol "(")  (symbol ")")
   (symbol "#{") (symbol "}")
   (symbol "{")  (symbol "}")})

(defn- add-edn-path-postfix
  [path]
  (let [path          (vec path)
        extended-path (loop [elems path, postfix nil]
                        (if-not (seq elems)
                          (into path postfix)
                          (if-let [s (edn-complementary-symbol (first elems))]
                            (recur (rest elems) (cons s postfix))
                            (recur (rest elems) postfix))))]
    extended-path))

(defn print-edn-paths
  [edn]
  (let [paths (tree->associative-paths edn edn-adjs edn-show)
        paths (sort-edn-paths paths)
        paths (map add-edn-path-postfix paths)]

    (doseq [p paths] (println (str/join " " p)))))

;; (PRETTY) PRINTING DIFFS
(defn pprint-ddiff
  [x y]
  (ddiff/pretty-print (ddiff/diff x y)))
